#include <misc.h>
#include <params.h>
#define PCWDETRAIN

subroutine tphysbc (ztodt,   pblht,   tpert,   ts,      &
                    qpert,   precl,   precc,   precsl,  precsc,  &
                    asdir,   asdif,   aldir,   aldif,   snowh,   &
                    qrs,     qrl,     flwds,   fsns,    fsnt,    &
                    flns,    flnt,    lwup,    srfrad,  sols,    &
                    soll,    solsd,   solld,   cldo,    cldn,    &
                    tcwato,  tcwatn,  qcwato,  qcwatn,  lcwato,  &
                    lcwatn,  state,   tend,    icefrac, landfrac,&
		    ocnfrac, tin, prcsnw   )
!----------------------------------------------------------------------- 
! 
! Purpose: 
! Tendency physics BEFORE coupling to land, sea, and ice models.
! 
! Method: 
! Call physics subroutines and compute the following:
!     o cloud calculations (cloud fraction, emissivity, etc.)
!     o radiation calculations
! Pass surface fields for separate surface flux calculations
! Dump appropriate fields to history file.
! 
! Author: CCM1, CMS Contact: J. Truesdale
! 
!-----------------------------------------------------------------------

   use precision
   use ppgrid
   use phys_grid,     only: get_rlat_all_p, get_rlon_all_p
   use cldwat,        only: pcond
   use geopotential,  only: geopotential_t
   use physics_types, only: physics_state, physics_tend, physics_ptend, physics_update, physics_ptend_init
   use diagnostics,   only: diag_dynvar
   use history,       only: outfld
   use physconst,     only: gravit, latvap, cpair, tmelt, cappa, zvir, rair, rga
   use radheat,       only: radheat_net
   use constituents,  only: pcnst, pnats, ppcnst, qmin
   use tracers,       only: dcconnam, ixcldw
   use zm_conv,       only: zm_conv_evap, zm_convr
   use time_manager,  only: is_first_step, get_nstep, get_curr_calday
   use moistconvection, only: cmfmca

   implicit none

#include <comctl.h>
!
! Arguments
!
   real(r8), intent(in) :: ztodt                          ! 2 delta t (model time increment)
   real(r8), intent(in) :: ts(pcols)                      ! surface temperature
   real(r8), intent(in) :: tcwato(pcols,pver)             !cloud water old temperature
   real(r8), intent(in) :: qcwato(pcols,pver)             ! cloud water old q
   real(r8), intent(in) :: lcwato(pcols,pver)             ! cloud liquid water old q
   real(r8), intent(in) :: icefrac(pcols)                 ! sea ice fraction (fraction)
   real(r8), intent(in) :: landfrac(pcols)                ! land fraction (fraction)
   real(r8), intent(in) :: ocnfrac(pcols)                 ! ocean fraction (fraction)
   real(r8), intent(inout) :: pblht(pcols)                ! Planetary boundary layer height
   real(r8), intent(inout) :: tpert(pcols)                ! Thermal temperature excess
   real(r8), intent(inout) :: qpert(pcols,ppcnst)         ! Thermal humidity & constituent excess
   real(r8), intent(in) :: asdir(pcols)                  ! Albedo: shortwave, direct
   real(r8), intent(in) :: asdif(pcols)                  ! Albedo: shortwave, diffuse
   real(r8), intent(in) :: aldir(pcols)                  ! Albedo: longwave, direct
   real(r8), intent(in) :: aldif(pcols)                  ! Albedo: longwave, diffuse
   real(r8), intent(in) :: snowh(pcols)                  ! Snow depth (liquid water equivalent)
   real(r8), intent(inout) :: qrs(pcols,pver)            ! Shortwave heating rate
   real(r8), intent(inout) :: qrl(pcols,pver)            ! Longwave  heating rate
   real(r8), intent(inout) :: flwds(pcols)               ! Surface longwave down flux
   real(r8), intent(inout) :: fsns(pcols)                   ! Surface solar absorbed flux
   real(r8), intent(inout) :: fsnt(pcols)                   ! Net column abs solar flux at model top
   real(r8), intent(inout) :: flns(pcols)                   ! Srf longwave cooling (up-down) flux
   real(r8), intent(inout) :: flnt(pcols)                   ! Net outgoing lw flux at model top
   real(r8), intent(in) :: lwup(pcols)                    ! Surface longwave up flux
   real(r8), intent(out) :: srfrad(pcols)                 ! Net surface radiative flux (watts/m**2)
   real(r8), intent(inout) :: sols(pcols)                   ! Direct beam solar rad. onto srf (sw)
   real(r8), intent(inout) :: soll(pcols)                   ! Direct beam solar rad. onto srf (lw)
   real(r8), intent(inout) :: solsd(pcols)                  ! Diffuse solar radiation onto srf (sw)
   real(r8), intent(inout) :: solld(pcols)                  ! Diffuse solar radiation onto srf (lw)
   real(r8), intent(out) :: precl(pcols)                  ! Large-scale precipitation rate
   real(r8), intent(out) :: precc(pcols)                  ! Convective-scale preciptn rate
   real(r8), intent(out) :: precsl(pcols)                 ! L.S. snowfall rate
   real(r8), intent(out) :: precsc(pcols)                 ! C.S. snowfall rate
   real(r8), intent(inout) :: cldo(pcols,pver)            !old cloud fraction
   real(r8), intent(out) :: cldn(pcols,pver)              !new cloud fraction
   real(r8), intent(out) :: tcwatn(pcols,pver)            !cloud water new temperature
   real(r8), intent(out) :: qcwatn(pcols,pver)            ! cloud water new q
   real(r8), intent(out) :: lcwatn(pcols,pver)            ! cloud liq. water new q
   real(r8), intent(out) :: tin(pcols,pver)               ! input T, to compute FV output T
   real(r8), intent(out) :: prcsnw(pcols)                 ! snowfall rate (precsl + precsc)

   type(physics_state), intent(inout) :: state
   type(physics_tend ), intent(inout) :: tend
!
!---------------------------Local workspace-----------------------------
!
   real(r8) :: rhdfda(pcols,pver)            ! dRh/dcloud, old 
   real(r8) :: rhu00 (pcols,pver)            ! Rh threshold for cloud, old

   type(physics_ptend)   :: ptend                  ! indivdual parameterization tendencies

   integer :: nstep                          ! current timestep number

   real(r8) :: calday                        ! current calendar day
   real(r8) :: clat(pcols)                   ! current latitudes(radians)
   real(r8) :: clon(pcols)                   ! current longitudes(radians)

   real(r8) :: zdu(pcols,pver)               ! detraining mass flux from deep convection
   real(r8) :: ftem(pcols,pver)              ! Temporary workspace for outfld variables
   real(r8) :: cmfdqr(pcols,pver)            ! dq/dt due to moist convective rainout
   real(r8) :: cmfmc(pcols,pverp)            ! Convective mass flux--m sub c
   real(r8) :: cmfsl(pcols,pver)             ! Moist convection lw stat energy flux
   real(r8) :: cmflq(pcols,pver)             ! Moist convection total water flux
   real(r8) :: dtcond(pcols,pver)            ! dT/dt due to moist processes
   real(r8) :: dqcond(pcols,pver,ppcnst)     ! dq/dt due to moist processes

   real(r8) cldst(pcols,pver)
   real(r8) cltot(pcols)                      ! Diagnostic total cloud cover
   real(r8) cllow(pcols)                      !       "     low  cloud cover
   real(r8) clmed(pcols)                      !       "     mid  cloud cover
   real(r8) clhgh(pcols)                      !       "     hgh  cloud cover
   real(r8) cmfcme(pcols,pver)                ! cmf condensation - evaporation
   real(r8) cmfdqr2(pcols,pver)               ! dq/dt due to moist convective rainout
   real(r8) cmfmc2(pcols,pver)                ! Moist convection cloud mass flux
   real(r8) cmfsl2(pcols,pver)                ! Moist convection lw stat energy flux
   real(r8) cmflq2(pcols,pver)                ! Moist convection total water flux
   real(r8) cnt(pcols)                        ! Top level of convective activity
   real(r8) cnb(pcols)                        ! Lowest level of convective activity
   real(r8) cnt2(pcols)                       ! Top level of convective activity
   real(r8) cnb2(pcols)                       ! Bottom level of convective activity
   real(r8) concld(pcols,pver)             
   real(r8) coszrs(pcols)                     ! Cosine solar zenith angle
   real(r8) dlf(pcols,pver)                   ! Detraining cld H20 from convection
   real(r8) fwaut(pcols,pver)              
   real(r8) fsaut(pcols,pver)              
   real(r8) fracw(pcols,pver)              
   real(r8) fsacw(pcols,pver)              
   real(r8) fsaci(pcols,pver)              
   real(r8) nevapr(pcols,pver)                ! local evaporation of precipitation
   real(r8) prain(pcols,pver)                 ! local formation of precipitation
   real(r8) pflx(pcols,pverp)                 ! Conv rain flux thru out btm of lev
   real(r8) precc2(pcols)                     ! Convective-scale preciptn rate
   real(r8) preclp(pcols)                     ! sfc flux of precip from pcond
   real(r8) prect(pcols)                      ! total (conv+large scale) precip rate
   real(r8) qc(pcols,pver)                    ! dq/dt due to rainout terms
   real(r8) qc2(pcols,pver)                   ! dq/dt due to rainout terms
   real(r8) qme(pcols,pver)                   ! local condensation of cloud water
   real(r8) qpert2(pcols,ppcnst)              ! Perturbation q
   real(r8) rtdt                              ! 1./ztodt
   real(r8) tpert2(pcols)                     ! Perturbation T
   real(r8) tvm(pcols,pver)                   ! Virtual temperature
   real(r8) pmxrgn(pcols,pverp)               ! Maximum values of pressure for each
!                                             !    maximally overlapped region.
!                                             !    0->pmxrgn(i,1) is range of pressure for
!                                             !    1st region,pmxrgn(i,1)->pmxrgn(i,2) for
!                                             !    2nd region, etc
   integer lchnk                              ! chunk identifier
   integer ncol                               ! number of atmospheric columns

   integer nmxrgn(pcols)                      ! Number of maximally overlapped regions
   integer  i,k,m                             ! Longitude, level, constituent indices
                                           
!  real(r8) engt                              ! Thermal   energy integral
!  real(r8) engk                              ! Kinetic   energy integral
!  real(r8) engp                              ! Potential energy integral
   real(r8) clwp(pcols,pver)                  ! Presribed cloud liq. h2o path
   real(r8) rel(pcols,pver)                   ! Liquid cloud particle effective radius
   real(r8) rei(pcols,pver)                   ! Ice effective drop size (microns)
   real(r8) fice(pcols,pver)                  ! Fractional ice content within cloud
   real(r8) effcld(pcols,pver)                ! Effective cloud=cld*emis
   real(r8) emis(pcols,pver)                  ! Cloud longwave emissivity
   real(r8) clc(pcols)                        ! Total convective cloud (cloud scheme)
   real(r8) qtend(pcols,pver)                 ! moisture tendencies
   real(r8) ttend(pcols,pver)                 ! temp tendencies
   real(r8) lctend(pcols,pver)                ! cloud liquid water tendencies
   real(r8) rmelt(pcols,pver)                 ! heating rate due to phase change of precip
   real(r8) zero(pcols,pverp)                 ! a dummy array
   real(r8) clwp2(pcols,pver)                 ! in-cloud cloud water path
   real(r8) gclwp2(pcols,pver)                ! grid-box cloud water path
   real(r8) tgcwp(pcols)                      ! Vertically integrated cloud water path
   real(r8) tgiwp(pcols)                      ! Vertically integrated ice water path
   real(r8) tglwp(pcols)                      ! Vertically integrated liquid water path
   real(r8) tpw(pcols)                        ! Total precipitable water
   real(r8) hl (pcols)                        ! Liquid water scale height
!
   real(r8) dellow(pcols)                     ! delta p for bottom three levels of model
   real(r8) tavg(pcols)                       ! mass weighted average temperature for 
!                                          
! Used for OUTFLD only                     
!                                          
   real(r8) icimr(pcols,pver)                 ! in cloud ice mixing ratio
   real(r8) icwmr(pcols,pver)                 ! in cloud water mixing ratio
   real(r8) cldv(pcols,pver)                  ! cloud volume (fraction) occupied by rain or cloud water
   real(r8) rain(pcols,pver)                  ! total precip mixing ratio
   real(r8) icwmr1(pcols,pver)                ! in cloud water mixing ration for zhang scheme
   real(r8) icwmr2(pcols,pver)                ! in cloud water mixing ration for hack scheme
   real(r8) fracis(pcols,pver,ppcnst)         ! fraction of transported species that are insoluble
   real(r8) evappct(pcols)                    ! Convective-scale preciptn rate
   real(r8) timestep(pcols)
!
!     Variables for doing deep convective transport outside of zm_convr
!
   real(r8) mu2(pcols,pver)
   real(r8) eu2(pcols,pver)
   real(r8) du2(pcols,pver)
   real(r8) md2(pcols,pver)
   real(r8) ed2(pcols,pver)
   real(r8) dp(pcols,pver)
   real(r8) dsubcld(pcols)

   integer jt(pcols)
   integer maxg(pcols)
   integer ideep(pcols)
   integer lengath
!
!-----------------------------------------------------------------------
!
   lchnk = state%lchnk
   ncol  = state%ncol

   nstep = get_nstep()
   calday = get_curr_calday()
!
! Output NSTEP for debugging
!
   timestep(:ncol) = nstep
   call outfld ('NSTEP   ',timestep, pcols, lchnk)
!
!*** BAB's FV kludge
!
   tin(:ncol,:pver) = state%t(:ncol,:pver)
!
! Convert mixing ratio of non-water tracers to mass fraction of total
! atmospheric mass (Overwrite non-water portions of q3m1).
!
   if (ppcnst > 1) then
      call mr2mf (lchnk, ncol, state%q)
   end if
!
! Set reciprocal of layer thickness
! Set physics tendencies to 0
!
   state%rpdel(:ncol,:pver) = 1./state%pdel(:ncol,:pver)
   tend %dTdt(:ncol,:pver)  = 0.
   tend %dudt(:ncol,:pver)  = 0.
   tend %dvdt(:ncol,:pver)  = 0.
!
! Compute initial geopotential heights and dry static energies
!
   call geopotential_t (state%lnpint, state%lnpmid  , state%pint,  &
                        state%pmid  , state%pdel    , state%rpdel, &
                        state%t     , state%q(1,1,1), rair , gravit , zvir   , &
                        state%zi    , state%zm      , ncol      )

   state%s(:ncol,:pver) = cpair*state%t(:ncol,:pver) + gravit*state%zm(:ncol,:pver)
   call physics_ptend_init (ptend) ! Initialize parameterization tendency structure
!
! Make sure that input tracers are all positive (probably unnecessary)
!
   call qneg3('TPHYSBCb',lchnk  ,ncol    ,pcols   ,pver    , &
              ppcnst,qmin  ,state%q )
!
! Setup q and t accumulation fields
!
   dqcond(:ncol,:,:) = state%q(:ncol,:,:)
   dtcond(:ncol,:)   = state%s(:ncol,:)
!
! Zero out precip and convective fields before accumulating terms
!
   precl (:ncol)   = 0.
   preclp(:ncol)   = 0.
   precc (:ncol)   = 0.
   precsl(:ncol)   = 0.
   precsc(:ncol)   = 0.
   qc    (:ncol,:) = 0.
   cmfdqr(:ncol,:) = 0.
   cmfmc (:ncol,:) = 0.
   cmfsl (:ncol,:) = 0.
   cmflq (:ncol,:) = 0.
   dqcond(ncol+1:pcols,:,:) = 0.
   dtcond(ncol+1:pcols,:)   = 0.

   fracis (:ncol,:,1:ppcnst) = 1.
!
!===================================================
! Dry adjustment
!===================================================

! Copy state info for input to dadadj
! This is a kludge, so that dadadj does not have to be correctly reformulated in dry static energy

   ptend%s(:ncol,:pver)   = state%t(:ncol,:pver)
   ptend%q(:ncol,:pver,1) = state%q(:ncol,:pver,1)

   call t_startf ('dadadj')
   call dadadj (lchnk, ncol, state%pmid,  state%pint,  state%pdel,  &
                ptend%s, ptend%q(1,1,1))
   ptend%name  = 'dadadj'
   ptend%ls    = .TRUE.
   ptend%lq(1) = .TRUE.
   ptend%s(:ncol,:)   = (ptend%s(:ncol,:)   - state%t(:ncol,:)  )/ztodt * cpair
   ptend%q(:ncol,:,1) = (ptend%q(:ncol,:,1) - state%q(:ncol,:,1))/ztodt
   call t_stopf ('dadadj')
   call physics_update (state, tend, ptend, ztodt)
!
!===================================================
! Moist convection
!===================================================
!
! Since the PBL doesn't pass constituent perturbations, they
! are zeroed here for input to the moist convection routine
!
   qpert(:ncol,2:ppcnst) = 0.0
!
! JR Set some arrays to zero @ nstep=0. Otherwise random junk off the heap
! or stack will be used in zm_convr
!
   if (is_first_step()) then
      pblht(:ncol)  = 0.
      tpert(:ncol)  = 0.
      cldo (:ncol,:)= 0.
   end if
!
! Begin with Zhang-McFarlane (1996) convection parameterization
!
   call t_startf ('zm_convr')
   call zm_convr( lchnk,    ncol, &
                  state%t,   state%q,    precc,   cnt,     cnb,      &
                  pblht,   state%zm, state%phis,    state%zi,   ptend%q(:,:,1),     &
                  ptend%s, state%pmid,   state%pint,  state%pdel,  ts,       &
                  .5*ztodt,cmfmc,    cmfcme,  nstep,             &
                  tpert,   dlf,      pflx,    zdu,     cmfdqr,   &
                  mu2,      md2,     du2,     eu2,     ed2,      &
                  dp,       dsubcld, jt,      maxg,    ideep,    &
                  lengath, icwmr1    )
   ptend%name  = 'zm_convr'
   ptend%ls    = .TRUE.
   ptend%lq(1) = .TRUE.

   ftem(:ncol,:pver) = ptend%s(:ncol,:pver)/cpair
   call outfld('ZMDT    ',ftem           ,pcols   ,lchnk   )
   call outfld('ZMDQ    ',ptend%q(1,1,1) ,pcols   ,lchnk   )
   call t_stopf('zm_convr')

   call physics_update(state, tend, ptend, ztodt)
!
! Evaporate some of the precip directly into the environment (Sundqvist)
!
   call zm_conv_evap(state, ptend, pflx, precc, cldo, ztodt, evappct)
   ptend%name  = 'zm_conv_evap'
   ptend%ls    = .TRUE.
   ptend%lq(1) = .TRUE.
   call outfld('EVAPPCT ',evappct,pcols,state%lchnk)
   call physics_update(state, tend, ptend, ztodt)
!
! Transport cloud water only
!
   ptend%name = 'convtran1'
   do m=2,ppcnst
      if (m == ixcldw) ptend%lq(m) = .true.
   end do
   call t_startf ('convtran1')
   call convtran (lchnk,                                        &
                  ptend%lq,state%q, ppcnst,  mu2,     md2,   &
                  du2,     eu2,     ed2,     dp,      dsubcld,  &
                  jt,      maxg,    ideep,   1,       lengath,  &
                  nstep,   fracis,  ptend%q   )
   call t_stopf ('convtran1')
   call physics_update (state, tend, ptend, ztodt)
!
! Convert mass flux from reported mb/s to kg/m^2/s
!
   cmfmc(:ncol,:pver) = cmfmc(:ncol,:pver) * 100./gravit
!
! Add production of rain by zm_convr to qc.  Added 1 to k-index of pflx
! at instruction of PJR
!
   do k=2,pver
      do i=1,ncol
         qc(i,k) = qc(i,k) + (pflx(i,k+1) - pflx(i,k))*gravit/state%pdel(i,k)
      end do
   end do
!
! Call Hack (1994) convection scheme to deal with shallow/mid-level convection
! Begin by zeroing local copies of mass flux, energy fluxes, etc.
!
   cmfmc2 (:ncol,:pver) = 0.
   cmfdqr2(:ncol,:pver) = 0.
   cmfsl2 (:ncol,:pver) = 0.
   cmflq2 (:ncol,:pver) = 0.
   qc2    (:ncol,:pver) = 0.
!
! At PJR's instruction, deleted kludge to get past a once in a lifetime
! problem in cmfmca's transport of liq water due to reliance on m=2 being
! hard-wired to cloud water--JR.  Put back in after run bombed.
!
   where (abs(state%q(:ncol,:pver,ixcldw)) < 1.e-36)
      state%q(:ncol,:pver,ixcldw) = 0.
   end where

   call t_startf('cmfmca')
   tpert2(:ncol  ) =0.
   qpert2(:ncol,:) = qpert(:ncol,:)  ! BAB Why is this not zero, if tpert2=0???
   call cmfmca (lchnk,   ncol, &
                nstep,   ztodt,   state%pmid,  state%pdel,   &
                state%rpdel,   state%zm,      tpert2,  qpert2,  state%phis,     &
                pblht,   state%t,   state%q,   ptend%s,   ptend%q,      &
                cmfmc2,  cmfdqr2, cmfsl2,  cmflq2,  precc2,   &
                qc2,     cnt2,    cnb2,    icwmr2   )
   ptend%name  = 'cmfmca'
   ptend%ls    = .TRUE.
   ptend%lq(:) = .TRUE.

   ftem(:ncol,:pver) = ptend%s(:ncol,:pver)/cpair
   call outfld('CMFDT   ',ftem          ,pcols   ,lchnk   )
   call outfld('CMFDQ   ',ptend%q(1,1,1),pcols   ,lchnk   )
   call t_stopf('cmfmca')
   call physics_update (state, tend, ptend, ztodt)
!
! Merge shallow/mid-level output with prior results from Zhang-McFarlane
!
   do i=1,ncol
      precc(i) = precc(i) + precc2(i)
      if (cnt2(i) < cnt(i)) cnt(i) = cnt2(i)
      if (cnb2(i) > cnb(i)) cnb(i) = cnb2(i)
   end do
!
   cmfmc(:ncol,:pver)  = cmfmc(:ncol,:pver)  + cmfmc2(:ncol,:pver)
   cmfdqr(:ncol,:pver) = cmfdqr(:ncol,:pver) + cmfdqr2(:ncol,:pver)
   cmfsl(:ncol,:pver)  = cmfsl(:ncol,:pver)  + cmfsl2(:ncol,:pver)
   cmflq(:ncol,:pver)  = cmflq(:ncol,:pver)  + cmflq2(:ncol,:pver)
   qc(:ncol,:pver)     = qc(:ncol,:pver)     + qc2(:ncol,:pver)

#ifndef PCWDETRAIN
!
! put the detraining cloud water into precip to conserve
! mass
!
   do k = 1,pver
      do i = 1,ncol
         precc(i) = precc(i) + dlf(i,k)*state%pdel(i,k)/(gravit*1000.)
      end do
   end do
#else
!
! put the detraining cloud water into the cloud and environment in
! proportion to the cloud fraction
!
   do k = 1,pver
      do i = 1,ncol
         ptend%q(i,k,1)      = dlf(i,k)*(1.-cldo(i,k))
         ptend%s(i,k)        =-dlf(i,k)*(1.-cldo(i,k))*latvap
         ptend%q(i,k,ixcldw) = dlf(i,k)*cldo(i,k)
      end do
   end do
   ptend%name  = 'pcwdetrain'
   ptend%ls    = .TRUE.
   ptend%lq(1) = .TRUE.
   ptend%lq(ixcldw) = .TRUE.
   call physics_update(state, tend, ptend, ztodt)
#endif
!
! cloud fraction after transport and convection,
! derive the relationship between rh and cld from 
! the employed cloud scheme
!
   call t_startf('cldnrh')
   call cldnrh(lchnk,   ncol,                                &
               state%pmid,    state%t,   state%q(1,1,1),   state%omega, &
               cnt,     cnb,     cldn,    clc,     state%pdel,   &
               cmfmc,   landfrac,snowh,   concld,  cldst,    &
               ts,      state%pint(1,pverp),       zdu,  ocnfrac, &
               rhdfda,   rhu00 )
   call t_stopf('cldnrh')
!
! calculate the tendencies for moisture, temperature and cloud fraction
!
   rtdt = 1./ztodt
   qtend(:ncol,:pver) = (state%q(:ncol,:pver,1)       - qcwato(:ncol,:pver))*rtdt
   ttend(:ncol,:pver) = (state%t(:ncol,:pver)         - tcwato(:ncol,:pver))*rtdt
   lctend(:ncol,:pver) = (state%q(:ncol,:pver,ixcldw) - lcwato(:ncol,:pver))*rtdt
!
! strat condensation via prognostic cloud water
! calculate tendencies
!
   call t_startf('pcond')
   zero(:ncol,:pverp) = 0.
   call pcond (lchnk,   ncol, &
               state%t,   ttend,   state%q(1,1,1), qtend,       state%omega,     &
               state%q(1,1,ixcldw),state%pmid,     state%pdel,  cldn,     &
               qme,     nevapr,    prain,          rmelt,    &
               ztodt,   zero,      fwaut,          fsaut,       fracw,    &
               fsacw,   fsaci,     lctend,         rhdfda,      rhu00, icefrac)
   call t_stopf('pcond')
!
   call outfld('FWAUT',fwaut, pcols,lchnk)
   call outfld('FSAUT',fsaut, pcols,lchnk)
   call outfld('FRACW',fracw, pcols,lchnk)
   call outfld('FSACW',fsacw, pcols,lchnk)
   call outfld('FSACI',fsaci, pcols,lchnk)
!
! make it interactive
!
   do k = 1,pver
      do i = 1,ncol
         ptend%s(i,k)        = (qme(i,k) - nevapr(i,k))*latvap + rmelt(i,k)
         ptend%q(i,k,1)      =-(qme(i,k) - nevapr(i,k))
         ptend%q(i,k,ixcldw) = (qme(i,k) - prain(i,k))
         preclp(i) = preclp(i) + (prain(i,k)-nevapr(i,k))*state%pdel(i,k)/gravit
      end do
   end do
   ptend%name  = 'pcond'
   ptend%ls    = .TRUE.
   ptend%lq(1) = .TRUE.
   ptend%lq(ixcldw) = .TRUE.
   call physics_update (state, tend, ptend, ztodt)
!
! save off q and t after cloud water
!
   do k=1,pver
      qcwatn(:ncol,k) = state%q(:ncol,k,1)
      tcwatn(:ncol,k) = state%t(:ncol,k)
      lcwatn(:ncol,k) = state%q(:ncol,k,ixcldw)
   end do
!
!     Convective transport of all trace species except water vapor and
!     cloud water done here because we need to do the scavenging first
!     to determine the interstitial fraction.
!
   ptend%name  = 'convtran2'
   do m=2,ppcnst
      if (m /= ixcldw) ptend%lq(m) = .true.
   end do
   call t_startf ('convtran2')
   call convtran (lchnk,                                           &
                  ptend%lq,state%q, ppcnst,     mu2,     md2,      &
                  du2,     eu2,     ed2,        dp,      dsubcld,  &
                  jt,      maxg,    ideep,      1,       lengath,  &
                  nstep,   fracis,  ptend%q)
   call t_stopf ('convtran2')

   call physics_update (state, tend, ptend, ztodt)

   call outfld('CMFDQR  ',cmfdqr, pcols,lchnk)
   call outfld('CLDST   ',cldst,  pcols,lchnk)
   call outfld('CNVCLD  ',clc,    pcols,lchnk)
   call outfld('CONCLD  ',concld, pcols,lchnk)
   call outfld('CME     ',qme,    pcols,lchnk)
   call outfld('PRAIN   ',prain,  pcols,lchnk)
   call outfld('EVAPR   ',nevapr, pcols,lchnk)
   call outfld('DQP     ',qc,      pcols,   lchnk  )
!
! and add the precip of pcond and cond together. Note
! since preclp is in kg/m2/s,  We need to renormalize to m/s
! like other precips
!
   precl(:ncol) = precl(:ncol) + preclp(:ncol)/1000.
!
! Compute rate of temperature change due to moist processes
!
   dtcond(:ncol,:) = (state%s(:ncol,:) - dtcond(:ncol,:))*rtdt
   call outfld('DTCOND  ',dtcond / cpair  ,pcols   ,lchnk   )
!
! Compute rate of constituent change due to moist processes
!
   dqcond(:ncol,:,:) = (state%q(:ncol,:,:) - dqcond(:ncol,:,:))*rtdt
   do m=1,ppcnst
      call outfld(dcconnam(m),dqcond(1,1,m),pcols   ,lchnk )
   end do
!
!===================================================
! Moist physical parameteriztions complete: 
! send dynamical variables, and derived variables to history file
!===================================================
!
   call diag_dynvar (lchnk, ncol, state)
!
!===================================================
! Radiation computations
!===================================================
!
! Cosine solar zenith angle for current time step
!
   call get_rlat_all_p(lchnk, ncol, clat)
   call get_rlon_all_p(lchnk, ncol, clon)
   call zenith (calday, clat, clon, coszrs, ncol)
!
! Compute liquid water paths
!
   tgcwp(:ncol) = 0.
   do k=1,pver
      do i = 1,ncol
         gclwp2(i,k) = state%q(i,k,ixcldw)*state%pdel(i,k)/gravit*1000.0 ! Grid box liquid water path.
         clwp2(i,k) = gclwp2(i,k) / max(0.01_r8,cldn(i,k))               ! In-cloud liquid water path.
         tgcwp(i) = tgcwp(i) + gclwp2(i,k)
      end do
   end do

   call outfld('GCLDLWP  ',gclwp2,  pcols,lchnk)
   call outfld('ICLDLWP  ',clwp2,   pcols,lchnk)
   call outfld('TGCLDCWP ',tgcwp,  pcols,lchnk)

   if (dosw .or. dolw) then
!
! Compute cloud properties for input to radiation
!
      call t_startf('cldint')
      call virtem (ncol, pcols, pver, state%t, state%q(1,1,1), zvir, tvm)
      call cldint (lchnk, ncol, state%pmid, state%t, state%q(1,1,1), &
                   state%pint, state%lnpint, state%lnpmid, tvm, state%zi,  &
                   cldn, clwp, emis, effcld, landfrac,                    &
                   rel, rei, fice, state%pdel, tpw,                  &
                   hl, state%ps, nmxrgn, pmxrgn, clwp2)
      call t_stopf('cldint')
!
! Dump cloud field information to history tape buffer (diagnostics)
!
      call outfld('CLOUD   ',cldn,  pcols,lchnk)
      call outfld('EFFCLD  ',effcld, pcols,lchnk)
      call outfld('LWSH    ',hl,     pcols,lchnk)
!
! Compute in cloud ice mixing ratio and in cloud liquid mixing ratio
!
      do k=1,pver
         do i = 1,ncol
            icimr(i,k) = state%q(i,k,ixcldw)*fice(i,k) / max(0.01_r8,cldn(i,k))
            icwmr(i,k) = state%q(i,k,ixcldw)*(1.-fice(i,k)) / max(0.01_r8,cldn(i,k))
         end do
      end do

      call outfld('ICIMR ',icimr,  pcols,lchnk)
      call outfld('ICWMR ',icwmr,  pcols,lchnk)
      call outfld('FICE  ',fice,   pcols,lchnk)
!
! Special diagnostic cloud water fields:
!
      call outfld('SETLWP  ',clwp,   pcols,lchnk)
!
! Output pure ice and pure liquid water paths
!
      tgiwp(:ncol) = 0.

      do k=1,pver
         tgiwp(:ncol) = tgiwp(:ncol) + gclwp2(:ncol,k)*fice(:ncol,k)
      end do

      tglwp(:ncol) = tgcwp(:ncol) - tgiwp(:ncol)

      call outfld ('TGCLDLWP',tglwp,  pcols,lchnk)
      call outfld ('TGCLDIWP',tgiwp,  pcols,lchnk)
!
! Complete radiation calculations
!
      call t_startf ('radctl')
      call radctl (lchnk, ncol, lwup, emis, state%pmid,             &
                   state%pint, state%lnpmid, state%lnpint, state%t, state%q,   &
                   cldn, clwp2, coszrs, asdir, asdif,               &
                   aldir, aldif, pmxrgn, nmxrgn, fsns, fsnt    ,flns    ,flnt    , &
                   qrs, qrl, flwds, rel, rei,                       &
                   fice, sols, soll, solsd, solld,                  &
                   landfrac, state%zm)
      call t_stopf ('radctl')
!
! Cloud cover diagnostics
! radctl can change pmxrgn and nmxrgn so cldsav needs to follow 
! radctl.
!
      call cldsav (lchnk, ncol, cldn, state%pmid, cltot, &
                   cllow, clmed, clhgh, nmxrgn, pmxrgn)
!
! Dump cloud field information to history tape buffer (diagnostics)
!
      call outfld('CLDTOT  ',cltot  ,pcols,lchnk)
      call outfld('CLDLOW  ',cllow  ,pcols,lchnk)
      call outfld('CLDMED  ',clmed  ,pcols,lchnk)
      call outfld('CLDHGH  ',clhgh  ,pcols,lchnk)
   end if
!
! Compute net flux (for use in SLD energy fixer; not used in other dyn cores)
! Since fsns, fsnt, flns, and flnt are in the buffer, array values will be carried across
! timesteps when the radiation code is not invoked.
!
   do i=1,ncol
      tend%flx_net(i) = fsnt(i) - fsns(i) - flnt(i) + flns(i)
   end do
!
! Compute net radiative heating
!
   call radheat_net (state, ptend, qrl, qrs)
!
! Add radiation tendencies to cummulative model tendencies and update profiles
!
   call physics_update(state, tend, ptend, ztodt)
!
! Compute net surface radiative flux for use by surface temperature code.
! Note that units have already been converted to mks in RADCTL.  Since
! fsns and flwds are in the buffer, array values will be carried across
! timesteps when the radiation code is not invoked.
!
   srfrad(:ncol) = fsns(:ncol) + flwds(:ncol)
   call outfld('SRFRAD  ',srfrad,pcols,lchnk)
!
! determine whether precipitation, prec, is frozen (snow) or not
! by taking the mass-weighted temperature of the bottom most layers.
!
!  determine mass weighted average temperature in the bottom three 
!  levels of the model (approximately the lowest 900 meters)
!
   dellow(:ncol) = 0.0
   tavg (:ncol) = 0.0
!
   do k=pver-2, pver
      dellow(:ncol) = dellow(:ncol) + state%pdel(:ncol,k)   
      tavg(:ncol) = tavg(:ncol) + state%t(:ncol,k )*state%pdel(:ncol,k)
   end do
!
   tavg (:ncol) = tavg(:ncol)/dellow(:ncol)
!
   where (tavg(:ncol) > (tmelt-2.0) )
      precsc(:ncol) = 0.
      precsl(:ncol) = 0.
   elsewhere
      precsc(:ncol) = precc(:ncol)
      precsl(:ncol) = precl(:ncol)
   end where
   prcsnw(:ncol) = precsc(:ncol) + precsl(:ncol)   ! total snowfall rate: needed by slab ocean model
!
! Save atmospheric fields to force surface models
!
   call srfxfer (lchnk, ncol, state%ps, state%u(1,pver), state%v(1,pver),    &
                 state%t(1,pver), state%q(1,pver,1), state%exner(1,pver), state%zm(1,pver), &
                    state%pmid,      &
                 state%rpdel(1,pver))

!---------------------------------------------------------------------------------------
! Save history variables. These should move to the appropriate parameterization interface
!---------------------------------------------------------------------------------------

   call outfld('CMFMC   ',cmfmc   ,pcols   ,lchnk   )
   call outfld('CMFSL   ',cmfsl   ,pcols   ,lchnk   )
   call outfld('CMFLQ   ',cmflq   ,pcols   ,lchnk   )
   call outfld('PRECL   ',precl   ,pcols   ,lchnk       )
   call outfld('PRECC   ',precc   ,pcols   ,lchnk       )
   call outfld('PRECSL  ',precsl  ,pcols   ,lchnk       )
   call outfld('PRECSC  ',precsc  ,pcols   ,lchnk       )
   
   prect(:ncol) = precc(:ncol) + precl(:ncol)
   call outfld('PRECT   ',prect   ,pcols   ,lchnk       )
   call outfld('PRECTMX ',prect   ,pcols   ,lchnk       )

#if ( defined COUP_CSM )
   call outfld('PRECLav ',precl   ,pcols   ,lchnk   )
   call outfld('PRECCav ',precc   ,pcols   ,lchnk   )
#endif
!     
! Compute heating rate for dtheta/dt 
!
   do k=1,pver
      do i=1,ncol
         ftem(i,k) = (qrs(i,k) + qrl(i,k))/cpair * (1.e5/state%pmid(i,k))**cappa
      end do
   end do
   call outfld('HR      ',ftem    ,pcols   ,lchnk   )
!
! Convert mass fractions of non-water tracers back to mixing ratios.
! (Overwrite non-water portions of q3m1).
!
   if (ppcnst > 1) then
      call mf2mr (lchnk, ncol, state%q)
   end if

   return
end subroutine tphysbc
